/* ---------------------------------------------------------------------------------------------
 *										R.e.e.f.
 * ---------------------------------------------------------------------------------------------
 *
 *	projeto: dHome
 *	autores: Adelmario Douglas Leite Cabral Junior - adlcj[at]cin.ufpe.br
 *			 Lamberto Augusto De Oliveira Neto - laon[at]cin.ufpe.br
 *			 Marcelo Bezerra de Amorim - mba3[at]cin.ufpe.br
 *			 Francisco de Farias Dantas - ffd[at]cin.ufpe.br	
 *
 *  mdulo: PLC
 *  ultima atualizao: 26/11/2005
*/


//#include <REG51.H>
#include <REG517A.H>

/*
 * Constantes para configurao da comunicao na serial
 */				  
#define CT_CLK_FREQ 		12000000

//********************************************************************************************
//Defines relativos a COM1
//********************************************************************************************
#define CT_BAUD 			9600
#define CT_RELOAD_VALUE 	( 1024 - ( ( CT_CLK_FREQ / (32) ) / CT_BAUD ) )

#define SM     0x80		// Modo de operacao da serial 1
#define REN1   0x10		// habilita interface serial 1
#define TI1    0x02		// Interrupcao de transmissao da serial 1
#define NotTI1 0xFD
#define RI1    0x01		// Interrupcao de recepcao da serial 1
#define NotRI1 0xFE
//********************************************************************************************

//********************************************************************************************
//Defines relativos a COM0
//********************************************************************************************
#define CT_BAUD0 			9600
#define RELOAD_VALUE0 	( 1024 - ( ( CT_CLK_FREQ / (64) ) / CT_BAUD0 ) )
#define SM0 0x40
#define REN0   0x10		// habilita interface serial 0

//********************************************************************************************



//********************************************************************************************
//Defines relativos a COM emulada
//********************************************************************************************
#define CT_BAUD_E 			1200
#define CT_TIMER_RELOAD0 	( 256 - ( ( CT_CLK_FREQ / (384) ) / CT_BAUD_E ) )
#define CT_START_SERIAL 0
#define CT_STOP_SERIAL 1
//********************************************************************************************

//Estados principais do Mdulo PLC
//#define CT_WAITING0 0					//Estado 'Desativado0'
//#define CT_WAITING1 1					//Estado 'Desativado1'
//#define CT_RUNNING 2					//Estado 'Funcionando'

//Defines gerais
//#define CT_START0 0x0F
//#define CT_START1 0xAA

//********************************************************************************************
//Defines referente mdulo PLC
//********************************************************************************************
#define CT_ACK 0x06 						// corresponde ao estado que envia todo o comando X10
#define CT_OK 0xFF							// corresponde ao estado que indica que o comando foi enviado com sucesso
#define CT_FETCH 0xFE						// corresponde ao estado que recebe os dados do servidor
#define CT_START 0x02						// corresponde ao estado que envia o cdigo de iniciar comando para o dispositivo X10
#define CT_NAK 0x15							// corresponde ao estado que indica que o cdigo no foi identificado
#define CT_CR 0x0D							// corresponde ao cdigo de retorno
#define CT_ON 0xFD							// corresponde ao estado que verifica se  para enviar comando ou desativar conexo 
#define CT_NOP 0x99							// corresponde ao estado para iniciar conexo com o controlador X10
#define CT_ERRO 0x97						// corresponde ao estado que indica algum erro ocorrido 
#define CT_send_x10_command 0x63			// corresponde ao cdigo X10 de enviar comando
//********************************************************************************************

//********************************************************************************************
//Defines relativos ao drive do Servidor
//********************************************************************************************
//Envia
#define CT_ATIVAR_COMUNICACAO 0xC1
#define CT_DESATIVAR_COMUNICACAO 0xCF
#define CT_REQUESITOS_STATUS 0xC2
#define CT_ENVIAR_COMANDO 0xC3
//Recebe
#define CT_OK_SERV 0x01
#define CT_STATUS_ATUAL 0x02
#define CT_ESPERANDO_CODIGO_CASA 0x03
#define CT_ESPERANDO_CODIGO_APARELHO_FUNCAO 0x04
#define CT_CONTROLADOR_EXTRA_CODIGO 0x05
#define CT_ESPERANDO_EXTRA_DADOS 0x06
#define CT_TRANSMISSAO_COMPLETA 0x07
#define CT_DADOS_RECEBIDOS 0x08
#define CT_ERRO_PLC 0x9
#define CT_ERRO_SERV 0x0F
//********************************************************************************************

//Bits de controle do handshake (a definir)
//sbit STROBLE = P2^0;					//Sinal Stroble da paralela
//sbit ACK_P = P2^1;						//Sinal ACK para a paralela

//********************************************************************************************
//Variveis do Controlador
//********************************************************************************************
//static unsigned char KEY;							// chave usada na mquina de estados
unsigned char extended[64];							// varivel que armazena os dados extendido recebidos do servidor
unsigned char house_code_s;							// varivel que armazena o cdigo da casa recebido do servidor
unsigned char unit_code_s;							// varivel que armazena o cdigo do dispositivo recebido do servidor
unsigned char function_code_s;						// varivel que armazena o cdigo da funo recebido do servidor
unsigned char extra_code_s;							// varivel que armazena a quantidade da dimerizao
unsigned char dados = '0';							// varivel que auxilia no envio do comando 
unsigned char tmp, tmp2, tmp3, tmp4, tmp5;			// variveis que auxiliam no envio do comando
unsigned char tmp6, tmp7, tmp8,tmp9, tmp10, tmp11;	// variveis que auxiliam no envio do comando

int extra_lenght = 0x00; 									//varivel que armazena a quantidade de bytes do cdigo extendido

enum {ON, OFF, FETCH, START, ACK, ERRO, OK, NAK};
//********************************************************************************************


//********************************************************************************************
//Variveis da serial emulada
//********************************************************************************************
enum {WAITING, TRANSMITING, FINALIZING};                     //Estados de Transmio de Dados via serial emulada
enum {WAITING_RECEIVE, RECEIVING, FINALIZING_RECEIVE};       //Estados da Recepo de Dados via serial emulada

sbit sTx = P4^0;
sbit sRx = P4^1;

bit GTI = 0;
bit GRI = 0;

bit send = 0;
bit receive = 0;

unsigned char time = 2;
unsigned char timeR= 1;	//Botei timeR pq num sabia se pode trasmitir e receber ao mesmo tempo

unsigned char n = 7;
unsigned char nR = 0x01;
unsigned char SNBUF;
unsigned char RCBUF;
//********************************************************************************************


/*void timer0ISR() interrupt 1 using 0{
	static unsigned char sendState = WAITING;
	static unsigned char receiveState = WAITING_RECEIVE;

	TR0 = 0;
	TL0 += CT_TIMER_RELOAD0;
	//TR0 = 1;

	if(send){

		switch(sendState){

			case WAITING:

				sTx = CT_START_SERIAL;

				if(time == 0){

					time = 2;
					sendState = TRANSMITING;
					break;
				}
				time--;
				break;

			case TRANSMITING:

				if(SNBUF & 0x01){
					sTx = 1;
				} else {
					sTx = 0;
				}

				if((n == 0) && (time == 0)){

					time = 2;
					sendState = FINALIZING;
					break;

				} else if(time == 0){

					SNBUF = SNBUF >> 1;
					time = 2;
					n--;
					break;

				}
				time--;
				break;

			case FINALIZING:

				sTx = CT_STOP_SERIAL;

				if(time == 0){

					GTI = 1;
					time = 2;
					sendState = WAITING;
					break;

				}
				time--;
				break;

		}
	}


	if(receive){

      switch(receiveState){

		case WAITING_RECEIVE :
		    //Se o valor de sRx for igual a CT_START_SERIAL o byte deve ser recebido
		    if(sRx == CT_START_SERIAL){
		    //Garante que a verificao ser feita duas vezes para garantir
				if (timeR == 0){
		    		receiveState = RECEIVING;
			  		timeR = 1;
			  		RCBUF = 0;
			  		GRI = 0;
				}
				else {
			  		timeR--;
				}
            } else { //Caso o CT_START_SERIAL for recebido apenas uma vez, pode ser simplesmente ruido
		    	timeR = 1;
		    }
		   break;

	    case RECEIVING :
            //Verifica se  a primeira vez que o bit est sendo analisado
		    if(timeR){
		    	//Se o bit recebido for 1 o bit correspondente no buffer ser setado
                if(sRx){
		 	    	RCBUF = RCBUF | nR;
				}
			//Se o bit recebido for 0, o valor que j era 0 no se altera
				timeR--;
		   }

		   //Na segunda verificao, no importa o que tem no bit de recepo
		  else {
          	timeR = 1;
		    nR = nR << 1;
		    //Caso nR seja igual a zero, todos os bits j foram verificados
		    if(!nR){
               	receiveState = FINALIZING_RECEIVE;
            }
          }

         break;

		case FINALIZING_RECEIVE :
        	if(sRx == CT_STOP_SERIAL){
				if(timeR){
		   	    	timeR--;
				} else {
			    	GRI = 1;
			    	timeR = 1;
			    	receiveState = WAITING_RECEIVE;
			    	nR = 00000001;
				}
		    } else {   // verificar se est correto depois
			    GRI = 0;
			    timeR = 1;
			    receiveState = WAITING_RECEIVE;
			    nR = 00000001;
            }
		   break;
       }
	}
	TR0 = 1;
}*/

/*
*Inicializa COM1 modo 1 full-duplex
*/

void com1_inicializaSerial()
   {
   S1RELH = CT_RELOAD_VALUE >> 8;
   S1RELL = CT_RELOAD_VALUE;
   S1CON = SM | REN1; // Ativa recepcao serial 1 no Modo B (8-bit), com stop bit.
   }

/*
*Inicializa COM0 modo 1 full-duplex
*/

void com0_inicializaSerial(){

	//S0RELH = RELOAD_VALUE0 >> 8;
    //S0RELL = RELOAD_VALUE0;
	PCON |= 0x80;   // SMOD = 1
	ADCON0 = (ADCON0 & 0x7F) | 0x80;//BD;				//Habilita uso do bauda rate generator
	S0CON = SM0 | REN0;
}

/*
 * Envia um dado pela serial
 *
 * msg - dado a ser enviado pela serial
 */

void sendToServer(unsigned char msg){
	S1BUF = msg;								// envia o dado
	while(!(S1CON & TI1));								// espera terminar de transmitir o dado
	S1CON &= NotTI1;									// resetando o TI
}


/*
 * Retorna um comando recebido do Servidor via serial
 */

unsigned char receiveFromServer(){
	unsigned char temp = 0;
	while(!(S1CON & RI1));					// checa se o dado foi completamente recebido
	temp = S1BUF;							// copia o dado recebido para temp
	S1CON &= NotRI1;
	return temp;
}

/*
 * Retorna um dado recebido dos dispositivos via Serial simulada 
 */

unsigned char receiveSerial(){
	unsigned char temp = 0;
	while(!RI0);			// checa se o dado foi completamente recebido
	temp = S0BUF;			// copia o dado recebido para temp
	RI0 = 0;
	return temp;
}

/*
 * Envia um dado pela Serial simulada
 *
 * msn - mensagem a ser enviada
 */
void sendSerial(unsigned char msn){
	S0BUF = msn;			// envia o dado
	while(!TI0);			// espera terminar de transmitir o dado
	TI0 = 0;
}



/*
*quina de estados que recebe comandos do Power Linc
*/
/*void receiveX10(){
	;									//Que cdigos ASCII o Power Linc manda pra que eu saiba que chegou um comando aleatorio???
}*/

/*
 * Intercala um bit negado de cada bit dos 4 bits menos significativos do byte passado como parmetro,
 * retornando um byte com os bits intercalados.
 *
 * src - byte com os bits menos significativos contendo o que se deseja intercalar com bits negados.
 */

unsigned char transformByte(unsigned char src){
	unsigned char ret, nsrc, and;
	int ind;
	
	nsrc = !src;							//byte que conter os bits negados a serem intercalados
	ret = 0x00;								//byte que ser retornado
	and = 0x80;							
	
    //lao que intercala os bits negados
	for(ind = 0; ind < 4; ind++){//incio do for
		ret = ret | ((src>>ind)&and);		//coloca um bit dos quatro bits originais recebidos
		and = and>>1;
		ret = ret | ((nsrc>>(ind+1))&and);	//coloca o respectivo bit negado
		and = and>>1;						
	}//fim do for

	return ret;
}


/*
 * Funo que envia o comando X10 requisitado pelo servidor
 *
 * house_code    - Cdigo da casa recebido do servidor
 * unit_code     - Cdigo do dispositivo a ser manipulado recebido do servidor 
 * function_code - Cdigo da funo a ser executada recebido do servidor
 */

void sendX10Command(unsigned char house_code, unsigned char unit_code, unsigned char function_code, unsigned char extra_code){
	unsigned char house_code_t, function_code_t, cnt, repeat;
	int ind;

	//verifica se a funo recebida  de dimerizao ou contm dados extendidos
	if( function_code != 0x59 ){//incio do if
		sendSerial(0x63);							//envia o comando padro de X10 
		sendSerial(house_code);						//envia o cdigo da casa
		sendSerial(unit_code);						//envia o cdigo do dispositivo
		sendSerial(function_code);					//envia a funo
		sendSerial(extra_code);					
				
		tmp2 = receiveSerial();	  	
		tmp3 = receiveSerial();
		tmp4 = receiveSerial();
		tmp5 = receiveSerial();
		tmp6 = receiveSerial();
		tmp7 = receiveSerial();
		tmp8 = receiveSerial();
		tmp9 = receiveSerial();
		tmp10 = receiveSerial();
		tmp11 = receiveSerial();
		

	    for(repeat = 5*extra_lenght; repeat > 0x05; repeat--){
			receiveSerial();			
		}	
		
		if(extra_code_s == 0x00){
			extra_code_s = 0x31;			
		}
		else {
			extra_code_s = extra_code_s - 0x10;	        
		}

	}//fim do if 
	else {//incio do else	     
		cnt = (house_code_s<<4);		
		house_code_t = transformByte(cnt);								//coloca os bits intercalados do house_code_s
		function_code_t = transformByte((function_code_s<<3) & 0xFE);   //coloca os bits intercalados do function_code_s
		
		sendSerial(0x80);												// envia o cdigo padro de dados extendidos
		sendSerial(extra_code_s);										// envia a qauntidade de bytes dos dados extendidos
		
		dados = 0xE0;													// CT_START code no nibble mais significativo
		dados = (dados | (house_code_t>>4));
		sendSerial(dados);												// CT_START_code + 4 bits mais significativos de house_code com bits barrados intercalados 
		
		dados = ((0x0F & house_code_t)<<4);								// 4 bits menos significativos de house_code com bits barrados intercalados 
		dados = (dados | (function_code_t>>4));
		sendSerial(dados); 												// 4 bits menos significativos de house_code com bits barrados intercalados + 4 bits mais significativos de function_code com bits barrados intercalados 
		
		dados = (((0x0F & function_code_t)<<4) & 0xF8); 				// 4 bits menos significativos de function_code com bits barrados intercalados + 10
		dados = (dados | (extended[0]>>6)); 
		sendSerial(dados); 												// 4 bits menos significativos de function_code com bits barrados intercalados + 4 bits menos significativos do primeiro byte do dado extendindo

		//lao para enviar os dados extendidos desde do 3 bit at restarem 6 bits
		for (ind = 0; ind < ((2*extra_lenght)-1); ind++){//incio do foror
			dados = ((extended[ind]<<2) & 0xFC);
			dados = (dados | ((extended[ind+1])>>6));
			sendSerial(dados);
		}//fim do for
		
		dados = ((0xFC & extended[ind])<<2);
		sendSerial(dados);												//envia os 6 bits restantes junto com 2 zeros dos 6 definidos no padro X10
		
		dados = 0x00;
		sendSerial(dados);												//envia os 4 zeros restantes do padro X10
	}//fim do else
}


/*
 * Verifica se os dados recebidos so vlidos
 */

unsigned char check(){
	unsigned char cnt;
	//Verifica se no contm dados extendidos
	if(function_code_s != 0x59){//incio do if		
		if(tmp2 != 0x58) cnt = CT_ERRO; 								//transmission received		
		if((cnt != CT_ERRO) && (tmp3 != house_code_s)) cnt = CT_ERRO;						   		
		if((cnt != CT_ERRO) && (tmp4 != unit_code_s)) cnt = CT_ERRO;						   		
//		if((cnt != CT_ERRO) && (tmp5 != extra_code_s)) cnt = CT_ERRO; 			//sent o nmero de vezes				   		
		if((cnt != CT_ERRO) && (tmp5 != 0x31)) cnt = CT_ERRO; 			//sent o nmero de vezes				   		
		if((cnt != CT_ERRO) && (tmp6 != CT_CR)) cnt = CT_ERRO;										   		
																									   
		if((cnt != CT_ERRO) && (tmp7 != 0x58)) cnt = CT_ERRO; 			//transmission received		
		if((cnt != CT_ERRO) && (tmp8 != house_code_s)) cnt = CT_ERRO;							   		
		if((cnt != CT_ERRO) && (tmp9 != function_code_s)) cnt = CT_ERRO;						   		
//		if((cnt != CT_ERRO) && (tmp10 != extra_code_s)) cnt = CT_ERRO; 			//sent once						   		
		if((cnt != CT_ERRO) && (tmp10 != 0x31)) cnt = CT_ERRO; 			//sent once						   		
		if((cnt != CT_ERRO) && (tmp11 != CT_CR)) cnt = CT_ERRO;
	}//fim do if
	return cnt;
}


/*
 * Funo principal que contm a mquina de estados
 */

void sendX10(){
	static unsigned char KEY = OFF;	
	unsigned char cod1, cod2;	
	int index;
	//unsigned char bool = b;
	//unsigned char data_tmp = dt;
	static unsigned int erro = 0;

			
	//mquina de estados do controlador
	switch(KEY) {//incio do switch
		case OFF://caso que verifica se h algo novo vindo do servidor
			erro = 1;

			dados = receiveFromServer();

			//verifica se  para ativar X10
			if (dados == CT_ATIVAR_COMUNICACAO) {//incio do if 
				sendToServer(CT_OK_SERV); 			//pe ok no barramento
		 		KEY = ON; 						//vai pra CT_ON
			}//fim do if 
			else sendToServer(CT_ERRO_SERV); 			//erro

			break;

		case ON://caso que envia comando ou destiva a conexo

			erro = 1;
			
			dados = receiveFromServer();

			// verifica se  para desativar conexo
			if(dados == CT_DESATIVAR_COMUNICACAO ){//incio do if 
				sendToServer(CT_OK_SERV); 			// pe CT_OK no barramento
		 		KEY = OFF; 						// vai pra CT_NOP
			}//fim do if
			//verifica se  para enviar comando
		 	else if(dados == CT_ENVIAR_COMANDO ){//incio do else if
			sendToServer(CT_ESPERANDO_CODIGO_CASA); 			//esperando o codigo da casa
				KEY = FETCH; 					//vai para CT_FETCH
			}//fim do else if					
			else sendToServer(CT_ERRO_SERV); 			//erro

			break;

		case FETCH://caso que organiza os cdigos a serem enviados que esto sendo recebidos do Servidor
			house_code_s = receiveFromServer();
			sendToServer(CT_ESPERANDO_CODIGO_APARELHO_FUNCAO); 				//envia -> esperando cd, func. casa

			//dados = receiveFromServer();

			unit_code_s = receiveFromServer();//((dados >>3 ) & 0x1E);

			function_code_s = receiveFromServer();//((dados & 0x0F) << 1) | 0x01;

			sendToServer(CT_CONTROLADOR_EXTRA_CODIGO); 				//envia -> esperando extra cd

			dados = receiveFromServer();

			extra_code_s = dados;
			extra_lenght = dados;					//transforma dados no tipo int

			//verifica se contm dados extendidos
			if( function_code_s == 0x59 ){//incio do if
				sendToServer(CT_ESPERANDO_EXTRA_DADOS); 			//envia -> esperando extra dados
				
				//lao para colocar no buffer os dados extendidos recebidos
				for(index=0;index<extra_lenght;index++){//incio do for
					dados = receiveFromServer();						
					*(&extended+(index*2)) = transformByte(dados);
					*(&extended+((index*2)+1)) = transformByte(dados<<4);						
					//verifica se no  o ultimo byte a receber
					if(index!=extra_lenght-1){//incio do if
						sendToServer(CT_ESPERANDO_EXTRA_DADOS); 	//envia -> esperando extra dados
					}//fim do if
				}//fim do for
			}//fim do if

			KEY = START;						//vai para CT_START
			break;

		case START://caso que recebe o cdigo do X10 
			sendSerial(CT_START);
			cod1 = receiveSerial();					
			
			if(cod1 == CT_ACK){			   
				cod2 = receiveSerial();				
				KEY = ACK;

			}
			else if(cod1 == CT_NAK) KEY = NAK;
			else KEY = ERRO;				
			break;

		case NAK://caso que envia para o servidor se houve erro
			//verifica se houve erro
			if(erro){//incio do if
				KEY = START;
				erro = 0;
			}//fim do if
			else{//incio do else
				sendToServer(CT_ERRO_SERV); 	//envia -> CT_ERRO
				KEY = ON;
			}//fim do else
			break;

		case ACK://caso que envia o comando X10 recebido
//			receiveSerial();						 					//recebe o CT_CR
			
			if(extra_code_s == 0x00){
				extra_code_s = 0x41;			
			}
			else {	
        	    extra_code_s = 0x40 + extra_code_s;		//envia o repeat command (padro de X10)
			}
			sendX10Command(house_code_s ,unit_code_s ,function_code_s, extra_code_s); // envia o comando X10
			//sendX10Command(0x46,0x4C,0x45);
						
			//verifica se deu erro
			if(check() == CT_ERRO) KEY = ERRO;							// checa se existe algum erro
			else KEY = OK;
			break;

		case ERRO://caso que trata o erro
			//verifica se houve erro
			/*if(erro){//incio do if
                KEY = START;
				erro = 0;
			}//fim do if*/
			//else{//incio do else
				sendToServer(CT_ERRO_SERV); 						//envia -> CT_ERRO
				KEY = ON;
			//}//fim do else
			break;

		case OK://caso que indica que a execuo ocorreu com sucesso
			sendToServer(CT_TRANSMISSAO_COMPLETA); 				//envia -> transmisso completa
			KEY = ON;
			break;
	}//fim do switch 
}

/*
*Loop principal, inicializa portas e chama
*funes de enviar X10 e receber X10
*/
void main(){
	
	com0_inicializaSerial();				//Inicializa COM0
	com1_inicializaSerial();				//Inicializa COM1
	EAL = 1;								//Ativa interrupes

	//Loop principal
	while(1){

		sendX10();		 				//Envia comandos para o Power Linc
		//receiveX10();					//Recebe comandos do Power Linc
	}

	//Cdigos para teste
	/*sendSerial(CT_START);
	receiveSerial();
	receiveSerial();
	sendX10Command(0x46, 0x4C, 0x49);*/
	//sendSerial(0x4A);
/*

	sendSerial(CT_START);
	receiveSerial();
	receiveSerial();
	sendX10Command(0x46, 0x4C, 0x4B);		
	*/

}
