#include <REG517A.H>

#define CT_CLK_FREQ 		12000000
#define CT_BAUD_E 			1200
#define CT_TIMER_RELOAD0 	( 256 - ( ( CT_CLK_FREQ / (384) ) / CT_BAUD_E ) )
#define CT_START_SERIAL 0
#define CT_STOP_SERIAL 1

enum {WAITING_TRANSMIT, TRANSMITING, FINALIZING_TRANSMIT};   //Estados de Transmio de Dados via serial emulada
enum {WAITING_RECEIVE, RECEIVING, FINALIZING_RECEIVE};       //Estados da Recepo de Dados via serial emulada

sbit sTx = P4^0;				//Pino de sada da serial
sbit sRx = P4^1;				//Pino de entrada da serial

bit GTI = 0;					//Bit que indica se um byte foi transmitido
bit GRI = 0;					//Bit que indica se um byte foi recebido

bit send = 0;					//Bit que indica que uma transmisso serial est acontecendo
bit receive = 0;				//Bit que indica que uma recepo serial est acontecendo

unsigned char time = 1;			//Controla o tempo de transmisso da serial
unsigned char timeR= 1;			//Controla o tempo de recepo da serial

unsigned char nT = 0x07;		
unsigned char nR = 0x01;
unsigned char SNBUF;			//Buffer de envio
unsigned char RCBUF;			//Buffer de recepo

void timer0ISR() interrupt 1 using 0{
	static unsigned char sendState = WAITING_TRANSMIT;	//memoriza os estados de enviar serial
	static unsigned char receiveState = WAITING_RECEIVE;//memoriza os estados de receber serial

	TR0 = 0;					//Desliga serial
	TL0 += CT_TIMER_RELOAD0;	//Atualiza valor do timer
	//TR0 = 1;

	if(send){

		switch(sendState){
			//Transmite o start bit(0)
			case WAITING_TRANSMIT:

				sTx = CT_START_SERIAL;

				if(time == 0){

					time = 1;
					sendState = TRANSMITING;
					break;
				}
				time--;
				break;
			//Transmite os 8 bits de dados
			case TRANSMITING:

				if(SNBUF & 0x01){
					sTx = 1;
				} else {
					sTx = 0;
				}

				if((nT == 0) && (time == 0)){

					time = 1;
					sendState = FINALIZING_TRANSMIT;
					break;

				} else if(time == 0){

					SNBUF = SNBUF >> 1;
					time = 1;
					nT--;
					break;

				}
				time--;
				break;
			//Transmite o stop bit(1)
			case FINALIZING_TRANSMIT:

				sTx = CT_STOP_SERIAL;

				if(time == 0){

					GTI = 1;
					send = 0;//No sei...
					time = 1;
					sendState = WAITING_TRANSMIT;
					break;

				}
				time--;
				break;

		}
	}


	if(receive){

      switch(receiveState){

		case WAITING_RECEIVE :
		    //Se o valor de sRx for igual a CT_START_SERIAL o byte deve ser recebido
		    if(sRx == CT_START_SERIAL){
		    //Garante que a verificao ser feita duas vezes para garantir
				if (timeR == 0){
		    		receiveState = RECEIVING;
			  		timeR = 1;
			  		RCBUF = 0;
			  		//GRI = 0;
				}
				else {
			  		timeR--;
				}
            } else { //Caso o CT_START_SERIAL for recebido apenas uma vez, pode ser simplesmente ruido
		    	timeR = 1;
		    }
		   break;

	    case RECEIVING :
            //Verifica se  a primeira vez que o bit est sendo analisado
		    if(timeR){
		    	//Se o bit recebido for 1 o bit correspondente no buffer ser setado
                if(sRx){
		 	    	RCBUF = RCBUF | nR;
				}
			//Se o bit recebido for 0, o valor que j era 0 no se altera
				timeR--;
		   }

		   //Na segunda verificao, no importa o que tem no bit de recepo
		  else {
          	timeR = 1;
		    nR = nR << 1;
		    //Caso nR seja igual a zero, todos os bits j foram verificados
		    if(!nR){
               	receiveState = FINALIZING_RECEIVE;
            }
          }

         break;

		case FINALIZING_RECEIVE :
        	if(sRx == CT_STOP_SERIAL){
				if(timeR){
		   	    	timeR--;
				} else {
			    	GRI = 1;
			    	timeR = 1;
			    	receiveState = WAITING_RECEIVE;
			    	nR = 0x01;
				}
		    } else {   // verificar se est correto depois
			    //GRI = 1;
				receive = 0;
			    timeR = 1;
			    receiveState = WAITING_RECEIVE;
			    nR = 0x01;
            }
		   break;
       }
	}
	TR0 = 1;
}

//Timer0 mode 2 auto-reload
void initializeTimer0() {
	TR0 = 0;
	TMOD = (TMOD & 0xF0) | 0x02;
	TH0 = CT_TIMER_RELOAD0;
	TL0 = CT_TIMER_RELOAD0;
	ET0 = 1;
	TR0 = 1;
}

unsigned char receiveSerial(){
	receive = 1;
	while(!GRI);
	//receive = 0;
	GRI = 0;
	return RCBUF;
}

void sendSerial(unsigned char msn){
	SNBUF = msn;
	send = 1;
	while(!GTI);
	//send = 0;
	GTI = 0;
}


void main(){
	char x = 0xAA;
	initializeTimer0();
	EAL = 1;

	while(1){
		//sendSerial(x);
		//x++;
		x = receiveSerial();
	}

}